"""Summary

Attributes:
    numpy_to_weld_type_mapping (TYPE): Description
"""

from weld.weldobject import *
import numpy as np
import os
import sys

import pkg_resources

numpy_to_weld_type_mapping = {
    'str': WeldVec(WeldChar()),
    'int32': WeldInt(),
    'int64': WeldLong(),
    'float32': WeldFloat(),
    'float64': WeldDouble(),
    'bool': WeldBit()
}


def to_shared_lib(name):
    """
    Returns the name with the platform dependent shared library extension.

    Args:
    name (TYPE): Description
    """
    if sys.platform.startswith('linux'):
        return name + ".so"
    elif sys.platform.startswith('darwin'):
        return name + ".dylib"
    else:
        sys.exit(1)


class NumPyEncoder(WeldObjectEncoder):
    """Summary

    Attributes:
        utils (TYPE): Description
    """

    def __init__(self):
        """Summary
        """
        lib = to_shared_lib("numpy_weld_convertor")
        lib_file = pkg_resources.resource_filename(__name__, lib)
        self.utils = ctypes.PyDLL(lib_file)

    def py_to_weld_type(self, obj):
        """Summary

        Args:
            obj (TYPE): Description

        Returns:
            TYPE: Description

        Raises:
            Exception: Description
        """
        if isinstance(obj, np.ndarray):
            dtype = str(obj.dtype)
            if dtype == 'int32':
                base = WeldInt()
            elif dtype == 'int64':
                base = WeldLong()
            elif dtype == 'float32':
                base = WeldFloat()
            elif dtype == 'float64':
                base = WeldDouble()
            elif dtype == 'bool':
                base = WeldBit()
            else:
                base = WeldVec(WeldChar())  # TODO: Fix this
            for i in xrange(obj.ndim):
                base = WeldVec(base)
        elif isinstance(obj, str):
            base = WeldVec(WeldChar())
        else:
            raise Exception("Invalid object type: unable to infer NVL type")
        return base

    def encode(self, obj, num_threads):
        """Converts Python object to Weld object.

        Args:
            obj: Python object that needs to be converted to Weld format

        Returns:
            Weld formatted object
        """
        if isinstance(obj, np.ndarray):
            if obj.ndim == 1 and obj.dtype == 'int32':
                numpy_to_weld = self.utils.numpy_to_weld_int_arr
            elif obj.ndim == 1 and obj.dtype == 'int64':
                numpy_to_weld = self.utils.numpy_to_weld_long_arr
            elif obj.ndim == 1 and obj.dtype == 'float64':
                numpy_to_weld = self.utils.numpy_to_weld_double_arr
            elif obj.ndim == 2 and obj.dtype == 'int32':
                numpy_to_weld = self.utils.numpy_to_weld_int_arr_arr
            elif obj.ndim == 2 and obj.dtype == 'int64':
                numpy_to_weld = self.utils.numpy_to_weld_long_arr_arr
            elif obj.ndim == 2 and obj.dtype == 'float64':
                numpy_to_weld = self.utils.numpy_to_weld_double_arr_arr
            elif obj.ndim == 1 and obj.dtype == 'bool':
                numpy_to_weld = self.utils.numpy_to_weld_bool_arr
            else:
                numpy_to_weld = self.utils.numpy_to_weld_char_arr_arr
        elif isinstance(obj, str):
            numpy_to_weld = self.utils.numpy_to_weld_char_arr
        else:
            raise Exception("Unable to encode; invalid object type")

        numpy_to_weld.restype = self.py_to_weld_type(obj).ctype_class
        numpy_to_weld.argtypes = [py_object]
        weld_vec = numpy_to_weld(obj, num_threads)
        return weld_vec


class NumPyDecoder(WeldObjectDecoder):
    """Summary

    Attributes:
        utils (TYPE): Description
    """

    def __init__(self):
        """Summary
        """
        lib = to_shared_lib("numpy_weld_convertor")
        lib_file = pkg_resources.resource_filename(__name__, lib)
        self.utils = ctypes.PyDLL(lib_file)

    def decode(self, obj, restype, raw_ptr=False):
        """Converts Weld object to Python object.

        Args:
            obj: Result of Weld computation that needs to be decoded
            restype: Type of Weld computation result
            raw_ptr: Boolean indicating whether obj needs to be extracted
                     from WeldValue or not

        Returns:
            Python object representing result of the Weld computation
        """
        if raw_ptr:
            data = obj
        else:
            data = cweld.WeldValue(obj).data()
        result = ctypes.cast(data, ctypes.POINTER(restype.ctype_class)).contents

        if restype == WeldInt():
            data = cweld.WeldValue(obj).data()
            result = ctypes.cast(data, ctypes.POINTER(c_int)).contents.value
            return result
        elif restype == WeldLong():
            data = cweld.WeldValue(obj).data()
            result = ctypes.cast(data, ctypes.POINTER(c_long)).contents.value
            return result
        elif restype == WeldFloat():
            data = cweld.WeldValue(obj).data()
            result = ctypes.cast(data, ctypes.POINTER(c_float)).contents.value
            return float(result)
        elif restype == WeldDouble():
            data = cweld.WeldValue(obj).data()
            result = ctypes.cast(data, ctypes.POINTER(c_double)).contents.value
            return float(result)

        # Obj is a WeldVec(WeldInt()).ctype_class, which is a subclass of
        # ctypes._structure
        if restype == WeldVec(WeldBit()):
            weld_to_numpy = self.utils.weld_to_numpy_bool_arr
        elif restype == WeldVec(WeldLong()):
            weld_to_numpy = self.utils.weld_to_numpy_long_arr
        elif restype == WeldVec(WeldDouble()):
            weld_to_numpy = self.utils.weld_to_numpy_double_arr
        elif restype == WeldVec(WeldVec(WeldChar())):
            weld_to_numpy = self.utils.weld_to_numpy_char_arr_arr
        elif restype == WeldVec(WeldVec(WeldInt())):
            weld_to_numpy = self.utils.weld_to_numpy_int_arr_arr
        elif restype == WeldVec(WeldVec(WeldLong())):
            weld_to_numpy = self.utils.weld_to_numpy_long_arr_arr
        elif restype == WeldVec(WeldVec(WeldDouble())):
            weld_to_numpy = self.utils.weld_to_numpy_double_arr_arr
        elif restype == WeldVec(WeldInt()):
            weld_to_numpy = self.utils.weld_to_numpy_int_arr
        elif isinstance(restype, WeldStruct):
            ret_vecs = []
            # Iterate through all fields in the struct, and recursively call
            # decode.
            for field_type in restype.field_types:
                ret_vec = self.decode(data, field_type, raw_ptr=True)
                data += sizeof(field_type.ctype_class())
                ret_vecs.append(ret_vec)
            return tuple(ret_vecs)
        else:
            raise Exception("Unable to decode; invalid return type")

        weld_to_numpy.restype = py_object
        weld_to_numpy.argtypes = [restype.ctype_class]

        ret_vec = weld_to_numpy(result)
        return ret_vec



import numpy as np

import numpy_weld_impl
from lazy_op import LazyOpResult
from weld.weldobject import *


class NumpyArrayWeld(LazyOpResult):
    """Summary

    Attributes:
        dim (TYPE): Description
        expr (TYPE): Description
        weld_type (TYPE): Description
    """

    def __init__(self, expr, weld_type, dim=1):
        """Summary

        Args:
            expr (TYPE): Description
            weld_type (TYPE): Description
            dim (int, optional): Description
        """
        self.expr = expr
        self.weld_type = weld_type
        self.dim = dim

    def __div__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, LazyOpResult):
            other = other.expr
        return NumpyArrayWeld(
            numpy_weld_impl.div(
                self.expr,
                other,
                self.weld_type
            ),
            self.weld_type
        )

    def sum(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return NumpyArrayWeld(
            numpy_weld_impl.aggr(
                self.expr,
                "+",
                0,
                self.weld_type
            ),
            self.weld_type,
            0
        )


def dot(matrix, vector):
    """
    Computes the dot product between a matrix and a vector.
    TODO: Make this more generic

    Args:
        matrix (TYPE): Description
        vector (TYPE): Description
    """
    matrix_weld_type = None
    vector_weld_type = None

    if isinstance(matrix, LazyOpResult):
        matrix_weld_type = matrix.weld_type
        matrix = matrix.expr
    elif isinstance(matrix, np.ndarray):
        matrix_weld_type = numpy_weld_impl.numpy_to_weld_type_mapping[
            str(matrix.dtype)]

    if isinstance(vector, LazyOpResult):
        vector_weld_type = vector.weld_type
        vector = vector.expr
    elif isinstance(vector, np.ndarray):
        vector_weld_type = numpy_weld_impl.numpy_to_weld_type_mapping[
            str(vector.dtype)]

    return NumpyArrayWeld(
        numpy_weld_impl.dot(
            matrix,
            vector,
            matrix_weld_type,
            vector_weld_type),
        WeldDouble())


def exp(vector):
    """
    Computes a per-element exponent of the passed-in vector.

    Args:
        vector (TYPE): Description
    """
    weld_type = None
    if isinstance(vector, LazyOpResult):
        weld_type = vector.weld_type
        vector = vector.expr
    elif isinstance(vector, np.ndarray):
        weld_type = numpy_weld_impl.numpy_to_weld_type_mapping[
            str(vector.dtype)]
    return NumpyArrayWeld(numpy_weld_impl.exp(vector, weld_type), WeldDouble())



"""Contains implementations for each ported operation in NumPy.

Attributes:
    decoder_ (TYPE): Description
    encoder_ (TYPE): Description
    norm_factor_id_ (int): Unique IDs given to norm factor literals
"""
from encoders import *
from weld.weldobject import *

encoder_ = NumPyEncoder()
decoder_ = NumPyDecoder()


def div(array, other, ty):
    """
    Normalizes the passed-in array by the passed in quantity.

    Args:
        array (WeldObject / Numpy.ndarray): Input array to normalize
        other (WeldObject / float): Normalization factor
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    other_var = None
    if isinstance(other, WeldObject):
        other_var = other.obj_id
        weld_obj.dependencies[other_var] = other
    else:
        other_var = ("%.2f" % other)

    weld_template = """
       map(
         %(array)s,
         |value| value / %(ty)s(%(other)s)
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var,
                                          "other": other_var,
                                          "ty": ty}
    return weld_obj


def aggr(array, op, initial_value, ty):
    """
    Computes the aggregate of elements in the array.

    Args:
        array (WeldObject / Numpy.ndarray): Input array to aggregate
        op (str): Op string used to aggregate the array (+ / *)
        initial_value (int): Initial value for aggregation
        ty (WeldType): Type of each element in the input array


    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
      result(
        for(
          %(array)s,
          merger[%(ty)s,%(op)s],
          |b, i, e| merge(b, e)
        )
      )
    """
    weld_obj.weld_code = weld_template % {
        "array": array_var, "ty": ty, "op": op}
    return weld_obj


def dot(matrix, vector, matrix_ty, vector_ty):
    """
    Computes the dot product between a matrix and a vector.

    Args:
        matrix (WeldObject / Numpy.ndarray): 2-d input matrix
        vector (WeldObject / Numpy.ndarray): 1-d input vector
        ty (WeldType): Type of each element in the input matrix and vector

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    matrix_var = weld_obj.update(matrix)
    if isinstance(matrix, WeldObject):
        matrix_var = matrix.obj_id
        weld_obj.dependencies[matrix_var] = matrix

    vector_var = weld_obj.update(vector)
    loopsize_annotation = ""
    if isinstance(vector, WeldObject):
        vector_var = vector.obj_id
        weld_obj.dependencies[vector_var] = vector
    if isinstance(vector, np.ndarray):
        loopsize_annotation = "@(loopsize: %dL)" % len(vector)

    weld_template = """
       map(
         %(matrix)s,
         |row: vec[%(matrix_ty)s]|
           result(
             %(loopsize_annotation)s
             for(
               result(
                 %(loopsize_annotation)s
                 for(
                   zip(row, %(vector)s),
                   appender,
                   |b2, i2, e2: {%(matrix_ty)s, %(vector_ty)s}|
                     merge(b2, f64(e2.$0 * %(matrix_ty)s(e2.$1)))
                 )
               ),
               merger[f64,+],
               |b, i, e| merge(b, e)
             )
           )
       )
    """
    weld_obj.weld_code = weld_template % {"matrix": matrix_var,
                                          "vector": vector_var,
                                          "matrix_ty": matrix_ty,
                                          "vector_ty": vector_ty,
                                          "loopsize_annotation": loopsize_annotation}
    return weld_obj


def exp(array, ty):
    """
    Computes the per-element exponenet of the passed-in array.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
       map(
         %(array)s,
         |ele: %(ty)s| exp(ele)
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var, "ty": ty}
    return weld_obj



"""Contains implementations for each ported operation in Pandas.

Attributes:
    decoder_ (NumPyEncoder): Description
    encoder_ (NumPyDecoder): Description
"""
from encoders import *
from weld.weldobject import *


encoder_ = NumPyEncoder()
decoder_ = NumPyDecoder()

def get_field(expr, field):
    """ Fetch a field from a struct expr

    """
    weld_obj = WeldObject(encoder_, decoder_)

    struct_var = weld_obj.update(expr)
    if isinstance(expr, WeldObject):
        struct_var = expr.obj_id
        weld_obj.dependencies[struct_var] = expr

    weld_template = """
      %(struct)s.$%(field)s
    """

    weld_obj.weld_code = weld_template % {"struct":struct_var,
                                          "field":field}
    return weld_obj

def unique(array, ty):
    """
    Returns a new array-of-arrays with all duplicate arrays removed.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
       map(
         tovec(
           result(
             for(
               map(
                 %(array)s,
                 |p: %(ty)s| {p,0}
               ),
               dictmerger[%(ty)s,i32,+],
               |b, i, e| merge(b,e)
             )
           )
         ),
         |p: {%(ty)s, i32}| p.$0
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var, "ty": ty}
    return weld_obj


def aggr(array, op, initial_value, ty):
    """
    Returns sum of elements in the array.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        op (str): Op string used to aggregate the array (+ / *)
        initial_value (int): Initial value for aggregation
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
      result(
        for(
          %(array)s,
          merger[%(ty)s,%(op)s],
          |b, i, e| merge(b, e)
        )
      )
    """
    weld_obj.weld_code = weld_template % {
        "array": array_var, "ty": ty, "op": op}
    return weld_obj


def mask(array, predicates, new_value, ty):
    """
    Returns a new array, with each element in the original array satisfying the
    passed-in predicate set to `new_value`

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        predicates (WeldObject / Numpy.ndarray<bool>): Predicate set
        new_value (WeldObject / Numpy.ndarray / str): mask value
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    predicates_var = weld_obj.update(predicates)
    if isinstance(predicates, WeldObject):
        predicates_var = predicates.obj_id
        weld_obj.dependencies[predicates_var] = predicates

    if str(ty).startswith("vec"):
        new_value_var = weld_obj.update(new_value)
        if isinstance(new_value, WeldObject):
            new_value_var = new_value.obj_id
            weld_obj.dependencies[new_value_var] = new_value
    else:
        new_value_var = "%s(%s)" % (ty, str(new_value))

    weld_template = """
       map(
         zip(%(array)s, %(predicates)s),
         |p: {%(ty)s, bool}| if (p.$1, %(new_value)s, p.$0)
       )
    """
    weld_obj.weld_code = weld_template % {
        "array": array_var,
        "predicates": predicates_var,
        "new_value": new_value_var,
        "ty": ty}

    return weld_obj


def filter(array, predicates, ty=None):
    """
    Returns a new array, with each element in the original array satisfying the
    passed-in predicate set to `new_value`

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        predicates (WeldObject / Numpy.ndarray<bool>): Predicate set
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    predicates_var = weld_obj.update(predicates)
    if isinstance(predicates, WeldObject):
        predicates_var = predicates.obj_id
        weld_obj.dependencies[predicates_var] = predicates

    weld_template = """
       result(
         for(
           zip(%(array)s, %(predicates)s),
           appender,
           |b, i, e| if (e.$1, merge(b, e.$0), b)
         )
       )
    """
    weld_obj.weld_code = weld_template % {
        "array": array_var,
        "predicates": predicates_var}

    return weld_obj

def pivot_filter(pivot_array, predicates, ty=None):
    """
    Returns a new array, with each element in the original array satisfying the
    passed-in predicate set to `new_value`

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        predicates (WeldObject / Numpy.ndarray<bool>): Predicate set
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_array_var = weld_obj.update(pivot_array)
    if isinstance(pivot_array, WeldObject):
        pivot_array_var = pivot_array.obj_id
        weld_obj.dependencies[pivot_array_var] = pivot_array

    predicates_var = weld_obj.update(predicates)
    if isinstance(predicates, WeldObject):
        predicates_var = predicates.obj_id
        weld_obj.dependencies[predicates_var] = predicates

    weld_template = """
    let index_filtered =
      result(
        for(
          zip(%(array)s.$0, %(predicates)s),
          appender,
          |b, i, e| if (e.$1, merge(b, e.$0), b)
        )
      );
    let pivot_filtered =
      map(
        %(array)s.$1,
        |x|
          result(
            for(
              zip(x, %(predicates)s),
              appender,
              |b, i, e| if (e.$1, merge(b, e.$0), b)
            )
          )
      );
    {index_filtered, pivot_filtered, %(array)s.$2}
    """
    weld_obj.weld_code = weld_template % {
        "array": pivot_array_var,
        "predicates": predicates_var}

    return weld_obj

def isin(array, predicates, ty):
    """ Checks if elements in array is also in predicates
    
    # TODO: better parameter naming

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        predicates (WeldObject / Numpy.ndarray<bool>): Predicate set
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    predicates_var = weld_obj.update(predicates)
    if isinstance(predicates, WeldObject):
        predicates_var = predicates.obj_id
        weld_obj.dependencies[predicates_var] = predicates
    weld_template = """
    let check_dict =
      result(
        for(
          map(
            %(predicate)s,
            |p: %(ty)s| {p,0}
          ),
        dictmerger[%(ty)s,i32,+],
        |b, i, e| merge(b,e)
        )
      );
      map(
        %(array)s,
        |x: %(ty)s| keyexists(check_dict, x)
      )
    """
    weld_obj.weld_code = weld_template % {
        "array": array_var,
        "predicate": predicates_var,
        "ty": ty}

    return weld_obj


def element_wise_op(array, other, op, ty):
    """
    Operation of series and other, element-wise (binary operator add)

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        other (WeldObject / Numpy.ndarray): Second Input array
        op (str): Op string used to compute element-wise operation (+ / *)
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    other_var = weld_obj.update(other)
    if isinstance(other, WeldObject):
        other_var = other.obj_id
        weld_obj.dependencies[other_var] = other

    weld_template = """
       map(
         zip(%(array)s, %(other)s),
         |a| a.$0 %(op)s a.$1
       )
    """

    weld_obj.weld_code = weld_template % {"array": array_var,
                                          "other": other_var,
                                          "ty": ty, "op": op}
    return weld_obj

def unzip_columns(expr, column_types):
    """
    Zip together multiple columns.

    Args:
        columns (WeldObject / Numpy.ndarray): lust of columns

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)
    column_appenders = []
    struct_fields = []
    result_fields = []
    for i, column_type in enumerate(column_types):
        column_appenders.append("appender[%s]" % column_type)
        struct_fields.append("merge(b.$%s, e.$%s)" % (i, i))
        result_fields.append("result(unzip_builder.$%s)" % i)
    appender_string = "{%s}" % ", ".join(column_appenders)
    struct_string = "{%s}" % ", ".join(struct_fields)
    result_string = "{%s}" % ", ".join(result_fields)
    expr_var = weld_obj.update(expr)
    if isinstance(expr, WeldObject):
        expr_var = expr.obj_id
        weld_obj.dependencies[expr_var] = expr

    weld_template = """
    let unzip_builder = for(
      %(expr)s,
      %(appenders)s,
      |b,i,e| %(struct_builder)s
    );
    %(result)s
    """

    weld_obj.weld_code = weld_template % {"expr": expr_var,
                                          "appenders": appender_string,
                                          "struct_builder": struct_string,
                                          "result": result_string}
    return weld_obj

def sort(expr, field = None, keytype=None, ascending=True):
    """
    Sorts the vector.
    If the field parameter is provided then the sort
    operators on a vector of structs where the sort key
    is the field of the struct.

    Args:
      expr (WeldObject)
      field (Int)
    """
    weld_obj = WeldObject(encoder_, decoder_)

    expr_var = weld_obj.update(expr)
    if isinstance(expr, WeldObject):
        expr_var = expr.obj_id
        weld_obj.dependencies[expr_var] = expr

    if field is not None:
        key_str = "x.$%s" % field
    else:
        key_str = "x"

    if not ascending:
        # The type is not necessarily f64.
        key_str = key_str + "* %s(-1)" % keytype

    weld_template = """
    sort(%(expr)s, |x| %(key)s)
    """
    weld_obj.weld_code = weld_template % {"expr":expr_var, "key":key_str}
    return weld_obj

def slice_vec(expr, start, stop):
    """
    Slices the vector.

    Args:
      expr (WeldObject)
      start (Long)
      stop (Long)
    """
    weld_obj = WeldObject(encoder_, decoder_)

    expr_var = weld_obj.update(expr)
    if isinstance(expr, WeldObject):
        expr_var = expr.obj_id
        weld_obj.dependencies[expr_var] = expr

    weld_template = """
    slice(%(expr)s, %(start)sL, %(stop)sL)
    """
    weld_obj.weld_code = weld_template % {"expr":expr_var,
                                          "start":start,
                                          "stop":stop}
    return weld_obj

def zip_columns(columns):
    """
    Zip together multiple columns.

    Args:
        columns (WeldObject / Numpy.ndarray): lust of columns

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)
    column_vars = []
    for column in columns:
        col_var = weld_obj.update(column)
        if isinstance(column, WeldObject):
            col_var = column.obj_id
            weld_obj.dependencies[col_var] = column
        column_vars.append(col_var)

    arrays = ", ".join(column_vars)

    weld_template = """
       result(
         for(
           zip(%(array)s),
           appender,
           |b, i, e| merge(b, e)
         )
       )
    """
    weld_obj.weld_code = weld_template % {
        "array": arrays,
    }

    return weld_obj

def compare(array, other, op, ty_str):
    """
    Performs passed-in comparison op between every element in the passed-in
    array and other, and returns an array of booleans.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        other (WeldObject / Numpy.ndarray): Second input array
        op (str): Op string used for element-wise comparison (== >= <= !=)
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    # Strings need to be encoded into vec[char] array.
    # Constants can be added directly to NVL snippet.
    if isinstance(other, str) or isinstance(other, WeldObject):
        other_var = weld_obj.update(other)
        if isinstance(other, WeldObject):
            other_var = tmp.obj_id
            weld_obj.dependencies[other_var] = other
    else:
        other_var = "%s(%s)" % (ty_str, str(other))

    weld_template = """
       map(
         %(array)s,
         |a: %(ty)s| a %(op)s %(other)s
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var,
                                          "other": other_var,
                                          "op": op, "ty": ty_str}

    return weld_obj


def slice(array, start, size, ty):
    """
    Returns a new array-of-arrays with each array truncated, starting at
    index `start` for `length` characters.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        start (int): starting index
        size (int): length to truncate at
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
       map(
         %(array)s,
         |array: %(ty)s| slice(array, %(start)dL, %(size)dL)
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var, "start": start,
                                          "ty": ty, "size": size}

    return weld_obj

def to_lower(array, ty):
    """
    Returns a new array of strings that are converted to lowercase

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        start (int): starting index
        size (int): length to truncate at
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
       map(
         %(array)s,
         |array: vec[i8]| map(array, |b:i8| if(b <= i8(90), b + i8(32), b))
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var, "ty": ty}

    return weld_obj

def contains(array, ty, string):
    """
    Checks if given string is contained in each string in the array.
    Output is a vec of booleans.

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        start (int): starting index
        size (int): length to truncate at
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    string_obj = weld_obj.update(string)
    if isinstance(string, WeldObject):
        string_obj =  string.obj_id
        weld_obj.dependencies[string_obj] = string

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    (start, end) = 0, len(string)
    # Some odd bug where iterating on str and slicing str results
    # in a segfault
    weld_template = """
       map(
         %(array)s,
         |str: vec[%(ty)s]|
           let tstr = str;
           result(
             for(
               str,
               merger[i8,+](i8(0)),
               |b,i,e|
                 if(slice(tstr, i, i64(%(end)s)) == %(cmpstr)s,
                    merge(b, i8(1)),
                    b
                 )
              )
            ) > i8(0)
       )
    """
    weld_obj.weld_code = weld_template % {"array": array_var, "ty": ty,
                                          "start": start, "end": end,
                                          "cmpstr": string_obj}

    return weld_obj


def count(array, ty):
    """
    Return number of non-NA/null observations in the Series
    TODO : Filter out NaN's once Weld supports NaN's

    Args:
        array (WeldObject / Numpy.ndarray): Input array
        ty (WeldType): Type of each element in the input array

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    array_var = weld_obj.update(array)
    if isinstance(array, WeldObject):
        array_var = array.obj_id
        weld_obj.dependencies[array_var] = array

    weld_template = """
     len(%(array)s)
  """

    weld_obj.weld_code = weld_template % {"array": array_var}

    return weld_obj

def join(expr1, expr2, d1_keys, d2_keys, keys_type, d1_vals, df1_vals_ty, d2_vals, df2_vals_ty):
    """
       Computes a join on two tables
    """
    weld_obj = WeldObject(encoder_, decoder_)

    df1_var = weld_obj.update(expr1)
    if isinstance(expr1, WeldObject):
        df1_var = expr1.obj_id
        weld_obj.dependencies[df1_var] = expr1

    df2_var = weld_obj.update(expr2)
    if isinstance(expr2, WeldObject):
        df2_var = expr2.obj_id
        weld_obj.dependencies[df2_var] = expr2

    d1_key_fields = ", ".join(["e.$%s" % k for k in d1_keys])
    if len(d1_keys) > 1:
        d1_key_struct = "{%s}" % (d1_key_fields)
    else:
        d1_key_struct = d1_key_fields

    d1_val_fields = ", ".join(["e.$%s" % k for k in d1_vals])

    d2_key_fields = ", ".join(["e.$%s" % k for k in d2_keys])
    if len(d2_keys) > 1:
        d2_key_struct = "{%s}" % (d2_key_fields)
    else:
        d2_key_struct = d2_key_fields

    d2_val_fields = ", ".join(["e.$%s" % k for k in d2_vals])
    d2_val_fields2 = ", ".join(["e2.$%s" % i for i, k in enumerate(d2_vals)])
    d2_val_struct = "{%s}" % (d2_val_fields)

    weld_template = """
    let df2_join_table = result(
      for(
        %(df2)s,
        groupmerger[%(kty)s, %(df2ty)s],
        |b, i, e| merge(b, {%(df2key)s, %(df2vals)s})
      )
    );

    result(for(
      %(df1)s,
      appender,
      |b, i, e|
        for(
          lookup(df2_join_table, %(df1key)s),
          b,
          |b2, i2, e2| merge(b, {%(df1key)s, %(df1vals)s, %(df2vals2)s})
        )
    ))
    """

    weld_obj.weld_code = weld_template % {"df1":df1_var,
                                          "df1key":d1_key_struct,
                                          "df1vals": d1_val_fields,
                                          "df2":df2_var,
                                          "kty":keys_type,
                                          "df2ty":df2_vals_ty,
                                          "df2key":d2_key_struct,
                                          "df2vals":d2_val_struct,
                                          "df2vals2":d2_val_fields2}
    
    return weld_obj

def pivot_table(expr, value_index, value_ty, index_index, index_ty, columns_index, columns_ty, aggfunc):
    """
    Constructs a pivot table where the index_index and columns_index are used as keys and the value_index is used as the value which is aggregated.
    """

    weld_obj = WeldObject(encoder_, decoder_)
    zip_var = weld_obj.update(expr)
    if isinstance(expr, WeldObject):
        zip_var = expr.obj_id
        weld_obj.dependencies[zip_var] = expr

    if aggfunc == 'sum':
        weld_template = """
          let bs = for(
            %(zip_expr)s,
            {dictmerger[{%(ity)s,%(cty)s},%(vty)s,+], dictmerger[%(ity)s,i64,+], dictmerger[%(cty)s,i64,+]},
            |b, i, e| {merge(b.$0, {{e.$%(id)s, e.$%(cd)s}, e.$%(vd)s}),
                       merge(b.$1, {e.$%(id)s, 1L}),
                       merge(b.$2, {e.$%(cd)s, 1L})}
          );
          let agg_dict = result(bs.$0);
          let ind_vec = sort(map(tovec(result(bs.$1)), |x| x.$0), |x| x);
          let col_vec = map(tovec(result(bs.$2)), |x| x.$0);
          let pivot = map(
            col_vec,
            |x:%(cty)s|
            map(ind_vec, |y:%(ity)s| f64(lookup(agg_dict, {y, x})))
        );
        {ind_vec, pivot, col_vec}
    """
    elif aggfunc == 'mean':
        weld_template = """
          let bs = for(
            %(zip_expr)s,
            {dictmerger[{%(ity)s,%(cty)s},{%(vty)s, i64},+], dictmerger[%(ity)s,i64,+], dictmerger[%(cty)s,i64,+]},
            |b, i, e| {merge(b.$0, {{e.$%(id)s, e.$%(cd)s}, {e.$%(vd)s, 1L}}),
                       merge(b.$1, {e.$%(id)s, 1L}),
                       merge(b.$2, {e.$%(cd)s, 1L})}
          );
          let agg_dict = result(bs.$0);
          let ind_vec = sort(map(tovec(result(bs.$1)), |x| x.$0), |x| x);
          let col_vec = map(tovec(result(bs.$2)), |x| x.$0);
          let pivot = map(
            col_vec,
            |x:%(cty)s|
            map(ind_vec, |y:%(ity)s| (
              let sum_len_pair = lookup(agg_dict, {y, x});
              f64(sum_len_pair.$0) / f64(sum_len_pair.$1))
            )
        );
        {ind_vec, pivot, col_vec}
    """

    else:
        raise Exception("Aggregate operation %s not supported." % aggfunc)

    weld_obj.weld_code = weld_template % {"zip_expr": zip_var,
                                          "ity": index_ty,
                                          "cty": columns_ty,
                                          "vty": value_ty,
                                          "id": index_index,
                                          "cd": columns_index,
                                          "vd": value_index}
    return weld_obj

def get_pivot_column(pivot, column_name, column_type):
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_var = weld_obj.update(pivot)
    if isinstance(pivot, WeldObject):
        pivot_var = pivot.obj_id
        weld_obj.dependencies[pivot_var] = pivot

    column_name_var = weld_obj.update(column_name)
    if isinstance(column_name, WeldObject):
        column_name_var = column_name.obj_id
        weld_obj.dependencies[column_name_var] = column_name

    weld_template = """
      let col_dict = result(for(
        %(piv)s.$2,
        dictmerger[%(cty)s,i64,+],
        |b, i, e| merge(b, {e, i})
      ));
    {%(piv)s.$0, lookup(%(piv)s.$1, lookup(col_dict, %(colnm)s))}
    """

    weld_obj.weld_code = weld_template % {"piv":pivot_var,
                                          "cty":column_type,
                                          "colnm":column_name_var}
    return weld_obj

def pivot_sort(pivot, column_name, index_type, column_type, pivot_type):
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_var = weld_obj.update(pivot)
    if isinstance(pivot, WeldObject):
        pivot_var = pivot.obj_id
        weld_obj.dependencies[pivot_var] = pivot

    column_name_var = weld_obj.update(column_name)
    if isinstance(column_name, WeldObject):
        column_name_var = column_name.obj_id
        weld_obj.dependencies[column_name_var] = column_name

        # Note the key_column is hardcoded for the movielens workload
        # diff is located at the 2nd index.
    weld_template = """
      let key_col = lookup(%(piv)s.$1, 2L);
      let sorted_indices = map(
        sort(
          result(
            for(
              key_col,
              appender[{%(pvty)s,i64}],
              |b,i,e| merge(b, {e,i})
            )
          ),
          |x| x.$0
        ),
        |x| x.$1
      );
      let new_piv = map(
        %(piv)s.$1,
        |x| result(
          for(
            x,
            appender[%(pvty)s],
            |b,i,e| merge(b, lookup(x, lookup(sorted_indices, i)))
          )
        )
      );
      let new_index = result(
        for(
          %(piv)s.$0,
          appender[%(idty)s],
          |b,i,e| merge(b, lookup(%(piv)s.$0, lookup(sorted_indices, i)))
        )
      );
    {new_index, new_piv, %(piv)s.$2}
    """

    weld_obj.weld_code = weld_template % {"piv": pivot_var,
                                          "cty": column_type,
                                          "idty": index_type,
                                          "colnm": column_name_var,
                                          "pvty": pivot_type}
    return weld_obj

def set_pivot_column(pivot, column_name, item, pivot_type, column_type):
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_var = weld_obj.update(pivot)
    if isinstance(pivot, WeldObject):
        pivot_var = pivot.obj_id
        weld_obj.dependencies[pivot_var] = pivot

    column_name_var = weld_obj.update(column_name)
    if isinstance(column_name, WeldObject):
        column_name_var = column_name.obj_id
        weld_obj.dependencies[column_name_var] = column_name

    item_var = weld_obj.update(item)
    if isinstance(item, WeldObject):
        item_var = item.obj_id
        weld_obj.dependencies[item_var] = item

    weld_template = """
    let pv_cols = for(
      %(piv)s.$1,
      appender[%(pvty)s],
      |b, i, e| merge(b, e)
    );
    let col_names = for(
      %(piv)s.$2,
      appender[%(cty)s],
      |b, i, e| merge(b, e)
    );
    let new_pivot_cols = result(merge(pv_cols, %(item)s));
    let new_col_names = result(merge(col_names, %(colnm)s));

    {%(piv)s.$0, new_pivot_cols, new_col_names}
    """

    weld_obj.weld_code = weld_template % {"piv":pivot_var,
                                          "cty":column_type,
                                          "pvty":pivot_type,
                                          "item":item_var,
                                          "colnm":column_name_var}
    return weld_obj

def pivot_sum(pivot, value_type):
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_var = weld_obj.update(pivot)
    if isinstance(pivot, WeldObject):
        pivot_var = pivot.obj_id
        weld_obj.dependencies[pivot_var] = pivot

    weld_template = """
      result(for(
        %(piv)s.$0,
        appender[%(pty)s],
        |b1,i1,e1|
          merge(b1,
            result(for(
              %(piv)s.$1,
              merger[%(pty)s,+],
              |b2, i2, e2| merge(b2, lookup(e2, i1))
            ))
          )
       ))
    """

    weld_obj.weld_code = weld_template % {"piv":pivot_var,
                                         "pty":value_type}
    return weld_obj

def pivot_div(pivot, series, vec_type, val_type):
    weld_obj = WeldObject(encoder_, decoder_)

    pivot_var = weld_obj.update(pivot)
    if isinstance(pivot, WeldObject):
        pivot_var = pivot.obj_id
        weld_obj.dependencies[pivot_var] = pivot

    series_var = weld_obj.update(series)
    if isinstance(series, WeldObject):
        series_var = series.obj_id
        weld_obj.dependencies[series_var] = series

    weld_template = """
    {%(piv)s.$0,
      map(
        %(piv)s.$1,
        |v:%(vty)s|
          result(for(
            v,
            appender[f64],
            |b, i, e| merge(b, f64(e) / f64(lookup(%(srs)s, i)))
          ))
      ),
     %(piv)s.$2
    }
    """

    weld_obj.weld_code = weld_template % {"piv":pivot_var,
                                          "pty":val_type,
                                          "vty":vec_type,
                                          "srs":series_var}
    return weld_obj

def groupby_sum(columns, column_tys, grouping_columns, grouping_column_tys):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    if len(grouping_columns) == 1 and len(grouping_column_tys) == 1:
        grouping_column_var = weld_obj.update(grouping_columns[0])
        if isinstance(grouping_columns[0], WeldObject):
            grouping_column_var = grouping_columns[0].weld_code
        grouping_column_ty_str = "%s" % grouping_column_tys[0]
    else:
        grouping_column_vars = []
        for column in grouping_columns:
            column_var = weld_obj.update(column)
            if isinstance(column_var, WeldObject):
                column_var = column_var.weld_code
            grouping_column_vars.append(column_var)
        grouping_column_var = ", ".join(grouping_column_vars)
        grouping_column_tys = [str(ty) for ty in grouping_column_tys]
        grouping_column_ty_str = ", ".join(grouping_column_tys)
        grouping_column_ty_str = "{%s}" % grouping_column_ty_str

    columns_var_list = []
    for column in columns:
        column_var = weld_obj.update(column)
        if isinstance(column, WeldObject):
            column_var = column.obj_id
            weld_obj.dependencies[column_var] = column
        columns_var_list.append(column_var)

    if len(columns_var_list) == 1 and len(grouping_columns) == 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        result_str = "merge(b, e)"
    elif len(columns_var_list) == 1 and len(grouping_columns) > 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        key_str_list = []
        for i in xrange(0, len(grouping_columns)):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str = "e.$" + str(len(grouping_columns))
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)
    elif len(columns_var_list) > 1 and len(grouping_columns) == 1:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str = "e.$0"
        value_str_list = []
        for i in xrange(1, len(columns) + 1):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)
    else:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str_list = []
        key_size = len(grouping_columns)
        value_size = len(columns)
        for i in xrange(0, key_size):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str_list = []
        for i in xrange(key_size, key_size + value_size):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)
    weld_template = """
     tovec(
       result(
         for(
           zip(%(grouping_column)s, %(columns)s),
           dictmerger[%(gty)s, %(ty)s, +],
           |b, i, e| %(result)s
         )
       )
     )
  """

    weld_obj.weld_code = weld_template % {"grouping_column": grouping_column_var,
                                          "columns": columns_var,
                                          "result": result_str,
                                          "ty": tys_str,
                                          "gty": grouping_column_ty_str}
    return weld_obj

def groupby_std(columns, column_tys, grouping_columns, grouping_column_tys):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    if len(grouping_columns) == 1 and len(grouping_column_tys) == 1:
        grouping_column_var = weld_obj.update(grouping_columns[0])
        if isinstance(grouping_columns[0], WeldObject):
            grouping_column_var = grouping_columns[0].weld_code
        grouping_column_ty_str = "%s" % grouping_column_tys[0]
    else:
        grouping_column_vars = []
        for column in grouping_columns:
            column_var = weld_obj.update(column)
            if isinstance(column_var, WeldObject):
                column_var = column_var.weld_code
            grouping_column_vars.append(column_var)
        grouping_column_var = ", ".join(grouping_column_vars)
        grouping_column_tys = [str(ty) for ty in grouping_column_tys]
        grouping_column_ty_str = ", ".join(grouping_column_tys)
        grouping_column_ty_str = "{%s}" % grouping_column_ty_str

    columns_var_list = []
    for column in columns:
        column_var = weld_obj.update(column)
        if isinstance(column, WeldObject):
            column_var = column.obj_id
            weld_obj.dependencies[column_var] = column
        columns_var_list.append(column_var)

    if len(columns_var_list) == 1 and len(grouping_columns) == 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        result_str = "merge(b, e)"
        # TODO : The above change needs to apply for all result strings
        # These currently break at the moment
    elif len(columns_var_list) == 1 and len(grouping_columns) > 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        key_str_list = []
        for i in xrange(0, len(grouping_columns)):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str = "e.$" + str(len(grouping_columns))
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s, 1L})" % ", ".join(result_str_list)
    elif len(columns_var_list) > 1 and len(grouping_columns) == 1:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str = "e.$0"
        value_str_list = []
        for i in xrange(1, len(columns) + 1):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, %s)" % ", ".join(result_str_list)
    else:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str_list = []
        key_size = len(grouping_columns)
        value_size = len(columns)
        for i in xrange(0, key_size):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str_list = []
        for i in xrange(key_size, key_size + value_size):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, %s)" % ", ".join(result_str_list)
        # TODO Need to implement exponent operator
        # The mean dict's e.$1.$0 assumes this is a scalar vector and not a struct vector
        # Also pandas normalizes by n - 1
    weld_template = """
    let sum_dict = result(
      for(
        zip(%(grouping_column)s, %(columns)s),
        groupmerger[%(gty)s, %(ty)s],
        |b, i, e| %(result)s
      )
    );
    map(
        sort(tovec(sum_dict), |x| x.$0),
        |e| {e.$0, 
                   (let sum = result(for(e.$1, merger[%(ty)s,+], |b,i,a| merge(b, a)));
                    let m = f64(sum) / f64(len(e.$1)); 
                    let msqr = result(for(e.$1, merger[f64,+], |b,i,a| merge(b, (f64(a)-m)*(f64(a)-m))));
                    sqrt(msqr / f64(len(e.$1) - 1L)) 
                   )}
    )
  """


    weld_obj.weld_code = weld_template % {"grouping_column": grouping_column_var,
                                          "columns": columns_var,
                                          "result": result_str,
                                          "ty": tys_str,
                                          "gty": grouping_column_ty_str}
    return weld_obj

def groupby_size(columns, column_tys, grouping_columns, grouping_column_tys):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    if len(grouping_columns) == 1 and len(grouping_column_tys) == 1:
        grouping_column_var = weld_obj.update(grouping_columns[0])
        if isinstance(grouping_columns[0], WeldObject):
            grouping_column_var = grouping_columns[0].weld_code
        grouping_column_ty_str = "%s" % grouping_column_tys[0]
    else:
        grouping_column_vars = []
        for column in grouping_columns:
            column_var = weld_obj.update(column)
            if isinstance(column_var, WeldObject):
                column_var = column_var.weld_code
            grouping_column_vars.append(column_var)
        grouping_column_var = ", ".join(grouping_column_vars)
        grouping_column_var = "zip(%s)" % grouping_column_var
        grouping_column_tys = [str(ty) for ty in grouping_column_tys]
        grouping_column_ty_str = ", ".join(grouping_column_tys)
        grouping_column_ty_str = "{%s}" % grouping_column_ty_str

    weld_template = """
    let group = sort(
      tovec(
        result(
          for(
            %(grouping_column)s,
            dictmerger[%(gty)s, i64, +],
            |b, i, e| merge(b, {e, 1L})
          )
        )
      ),
      |x:{%(gty)s, i64}| x.$0
    );
    {
      map(
      group,
        |x:{%(gty)s,i64}| x.$0
      ),
      map(
        group,
        |x:{%(gty)s,i64}| x.$1
      )
    }
  """

    weld_obj.weld_code = weld_template % {"grouping_column": grouping_column_var,
                                          "gty": grouping_column_ty_str}
    return weld_obj

def groupby_sort(columns, column_tys, grouping_columns, grouping_column_tys, key_index, ascending):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)
    if len(grouping_columns) == 1 and len(grouping_column_tys) == 1:
        grouping_column_var = weld_obj.update(grouping_columns[0])
        if isinstance(grouping_columns[0], WeldObject):
            grouping_column_var = grouping_columns[0].weld_code
        grouping_column_ty_str = "%s" % grouping_column_tys[0]
    else:
        grouping_column_vars = []
        for column in grouping_columns:
            column_var = weld_obj.update(column)
            if isinstance(column, WeldObject):
                column_var = column.obj_id
                weld_obj.dependencies[column_var] = column
            grouping_column_vars.append(column_var)
        grouping_column_var = ", ".join(grouping_column_vars)
        grouping_column_tys = [str(ty) for ty in grouping_column_tys]
        grouping_column_ty_str = ", ".join(grouping_column_tys)
        grouping_column_ty_str = "{%s}" % grouping_column_ty_str

    columns_var_list = []
    for column in columns:
        column_var = weld_obj.update(column)
        if isinstance(column, WeldObject):
            column_var = column.obj_id
            weld_obj.dependencies[column_var] = column
        columns_var_list.append(column_var)

    if len(columns_var_list) == 1 and len(grouping_columns) == 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        result_str = "merge(b, e)"
    elif len(columns_var_list) == 1 and len(grouping_columns) > 1:
        columns_var = columns_var_list[0]
        tys_str = column_tys[0]
        key_str_list = []
        for i in xrange(0, len(grouping_columns)):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str = "e.$" + str(len(grouping_columns))
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)
    elif len(columns_var_list) > 1 and len(grouping_columns) == 1:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str = "e.$0"
        value_str_list = []
        for i in xrange(1, len(columns) + 1):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)
    else:
        columns_var = "%s" % ", ".join(columns_var_list)
        column_tys = [str(ty) for ty in column_tys]
        tys_str = "{%s}" % ", ".join(column_tys)
        key_str_list = []
        key_size = len(grouping_columns)
        value_size = len(columns)
        for i in xrange(0, key_size):
            key_str_list.append("e.$%d" % i)
        key_str = "{%s}" % ", ".join(key_str_list)
        value_str_list = []
        for i in xrange(key_size, key_size + value_size):
            value_str_list.append("e.$%d" % i)
        value_str = "{%s}" % ", ".join(value_str_list)
        result_str_list = [key_str, value_str]
        result_str = "merge(b, {%s})" % ", ".join(result_str_list)

    if key_index == None:
        key_str = "x"
    else :
        key_str = "x.$%d" % key_index

    if ascending == False:
        key_str = key_str + "* %s(-1)" % column_tys[key_index]

    weld_template = """
    @(grain_size:1)map(
     tovec(
       result(
         for(
           zip(%(grouping_column)s, %(columns)s),
           groupmerger[%(gty)s, %(ty)s],
           |b, i, e| %(result)s
         )
       )
     ),
    |a:{%(gty)s, vec[%(ty)s]}| {a.$0, sort(a.$1, |x:%(ty)s| %(key_str)s)}
    )
  """

    weld_obj.weld_code = weld_template % {"grouping_column": grouping_column_var,
                                          "columns": columns_var,
                                          "result": result_str,
                                          "ty": tys_str,
                                          "gty": grouping_column_ty_str,
                                          "key_str": key_str}
    return weld_obj

def flatten_group(expr, column_tys, grouping_column_tys):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)

    group_var = weld_obj.update(expr)

    num_group_cols = len(grouping_column_tys)
    if num_group_cols == 1:
        grouping_column_ty_str = "%s" % grouping_column_tys[0]
        grouping_column_key_str = "a.$0"
    else:
        grouping_column_tys = [str(ty) for ty in grouping_column_tys]
        grouping_column_ty_str = ", ".join(grouping_column_tys)
        grouping_column_ty_str = "{%s}" % grouping_column_ty_str
        grouping_column_keys = ["a.$0.$%s" % i for i in range(0, num_group_cols)]
        grouping_column_key_str = "{%s}" % ", ".join(grouping_column_keys)

    if len(column_tys) == 1:
        tys_str = "%s" % column_tys[0]
        column_values_str = "b.$1"
    else:
        column_tys = [str(ty) for ty in column_tys]

        tys_str = "{%s}" % ", ".join(column_tys)
        column_values = ["b.$%s" % i for i in range(0, len(column_tys))]
        column_values_str = "{%s}" % ", ".join(column_values)

    # TODO: The output in pandas keps the keys sorted (even if keys are structs)
    # We need to allow keyfunctions in sort by clauses to be able to compare structs.
    # sort(%(group_vec)s, |x:{%(gty)s, vec[%(ty)s]}| x.$0),
    weld_template = """
    flatten(
      map(
        %(group_vec)s,
        |a:{%(gty)s, vec[%(ty)s]}| map(a.$1, |b:%(ty)s| {%(gkeys)s, %(gvals)s})
      )
    )
  """

    weld_obj.weld_code = weld_template % {"group_vec": expr.weld_code,
                                          "gkeys": grouping_column_key_str,
                                          "gvals": column_values_str,
                                          "ty": tys_str,
                                          "gty": grouping_column_ty_str}
    return weld_obj

def grouped_slice(expr, type, start, size):
    """
    Groups the given columns by the corresponding grouping column
    value, and aggregate by summing values.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        grouping_column (WeldObject): Column to group rest of columns by

    Returns:
          A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)
    weld_obj.update(expr)
    weld_template = """
       map(
           %(vec)s,
           |b: %(type)s| {b.$0, slice(b.$1, i64(%(start)s), i64(%(size)s))}
         )
    """

    weld_obj.weld_code = weld_template % {"vec": expr.weld_code,
                                          "type": type,
                                          "start": str(start),
                                          "size": str(size)}
    return weld_obj


def get_column(columns, column_tys, index):
    """
    Get column corresponding to passed-in index from ptr returned
    by groupBySum.

    Args:
        columns (List<WeldObject>): List of columns as WeldObjects
        column_tys (List<str>): List of each column data ty
        index (int): index of selected column

    Returns:
        A WeldObject representing this computation
    """
    weld_obj = WeldObject(encoder_, decoder_)
    columns_var = weld_obj.update(columns, tys=WeldVec(column_tys), override=False)
    if isinstance(columns, WeldObject):
        columns_var = columns.obj_id
        weld_obj.dependencies[columns_var] = columns

    weld_template = """
     map(
       %(columns)s,
       |elem: %(ty)s| elem.$%(index)s
     )
  """

    weld_obj.weld_code = weld_template % {"columns": columns_var,
                                          "ty": column_tys,
                                          "index": index}
    return weld_obj



"""Summary
"""
from weld.weldobject import *


def to_weld_type(weld_type, dim):
    """Summary

    Args:
        weld_type (TYPE): Description
        dim (TYPE): Description

    Returns:
        TYPE: Description
    """
    for i in xrange(dim):
        weld_type = WeldVec(weld_type)
    return weld_type


class LazyOpResult:
    """Wrapper class around as yet un-evaluated Weld computation results

    Attributes:
        dim (int): Dimensionality of the output
        expr (WeldObject / Numpy.ndarray): The expression that needs to be
            evaluated
        weld_type (WeldType): Type of the output object
    """

    def __init__(self, expr, weld_type, dim):
        """Summary

        Args:
            expr (TYPE): Description
            weld_type (TYPE): Description
            dim (TYPE): Description
        """
        self.expr = expr
        self.weld_type = weld_type
        self.dim = dim

    def evaluate(self, verbose=True, decode=True, passes=None, num_threads=1,
                 apply_experimental_transforms=False):
        """Summary

        Args:
            verbose (bool, optional): Description
            decode (bool, optional): Description

        Returns:
            TYPE: Description
        """
        if isinstance(self.expr, WeldObject):
            return self.expr.evaluate(
                to_weld_type(
                    self.weld_type,
                    self.dim),
                verbose,
                decode,
                passes=passes,
                num_threads=num_threads,
                apply_experimental_transforms=apply_experimental_transforms)
        return self.expr



import numpy as np
import pandas as pd

import grizzly_impl
from lazy_op import LazyOpResult, to_weld_type
from weld.weldobject import *

class DataFrameWeldLoc:
    """
    Label location based indexer for selection by label for dataframe objects.

    Attributes:
        df (TYPE): The DataFrame  being indexed into.
    """
    def __init__(self, df):
        self.df = df

    def __getitem__(self, key):
        if isinstance(key, SeriesWeld):
            # We're going to assume that the first column in these dataframes
            # is an index column. This assumption does not hold throughout grizzly,
            # so we should fix that moving forward.
            index_expr = grizzly_impl.get_field(self.df.expr, 0)
            if self.df.is_pivot:
                index_type, pivot_type, column_type = self.df.column_types
                index_elem_type = index_type.elemType
                index_expr_predicate = grizzly_impl.isin(index_expr, key.expr, index_elem_type)
                return DataFrameWeldExpr(
                    grizzly_impl.pivot_filter(
                        self.df.expr,
                        index_expr_predicate
                    ),
                    self.df.column_names,
                    self.df.weld_type,
                    is_pivot=True
                )
            # TODO : Need to implement for non-pivot tables
        raise Exception("Cannot invoke getitem on an object that is not SeriesWeld")

class WeldLocIndexer:
    """
    Label location based indexer for selection by label for Series objects.

    Attributes:
        grizzly_obj (TYPE): The Series being indexed into.
    """
    def __init__(self, grizzly_obj):
        # If index_type field of grizzly_obj is None
        # then we assume normal 0 - 1 indexing
        self.grizzly_obj = grizzly_obj

    def __getitem__(self, key):
        if isinstance(self.grizzly_obj, SeriesWeld):
            series = self.grizzly_obj
            if isinstance(key, SeriesWeld):
                if series.index_type is not None:
                    index_expr = grizzly_impl.get_field(series.expr, 0)
                    column_expr = grizzly_impl.get_field(series.expr, 1)
                    zip_expr = grizzly_impl.zip_columns([index_expr, column_expr])
                    predicate_expr = grizzly_impl.isin(index_expr, key.expr, series.index_type)
                    filtered_expr = grizzly_impl.filter(
                        zip_expr,
                        predicate_expr
                    )
                    unzip_expr = grizzly_impl.unzip_columns(
                        filtered_expr,
                        [series.index_type, series.weld_type]
                    )
                    return SeriesWeld(
                        unzip_expr,
                        series.weld_type,
                        series.df,
                        series.column_name,
                        series.index_type,
                        series.index_name
                    )
            # TODO : Need to implement for non-pivot tables
        raise Exception("Cannot invoke getitem on non SeriesWeld object")

class DataFrameWeldExpr:
    # TODO We need to merge this with the original DataFrameWeld class
    def __init__(self, expr, column_names, weld_type, is_pivot=False):
        if isinstance(weld_type, WeldStruct):
           self.column_types = weld_type.field_types
           self.weld_type = weld_type
        else:
            raise Exception("DataFrameWeldExpr can only except struct types")

        self.expr = expr
        self.column_names = column_names
        self.colindex_map = {name:i for i, name in enumerate(column_names)}
        self.is_pivot = is_pivot

    def __setitem__(self, key, item):
        if self.is_pivot:
            # Note that if this is a pivot table,
            # We have to modify the structure of the pivot table
            # that is append item to nested vector of vectors
            # and update the col_vec field
            # Also setitem appends a new item to the pivot table
            # Modifying an existing item is not implemented yet
            # TODO if pivot table check that the column being added
            # is same type
            if isinstance(item, SeriesWeld):
                if item.index_type is not None:
                    item_expr = grizzly_impl.get_field(item.expr, 1)
                else:
                    item_expr = item.expr
                self.expr = grizzly_impl.set_pivot_column(
                    self.expr,
                    key,
                    item_expr,
                    self.column_types[1].elemType,
                    self.column_types[2].elemType
                )
        else:
            raise Exception("Setitem not implemented for non-pivot table")

    def __getitem__(self, item):
        if self.is_pivot:
            # Note that if this is a pivot table,
            # then we can't check if the item is in the column
            # since we have not materialized the unique column values
            # Note if
            return SeriesWeld(
                grizzly_impl.get_pivot_column(
                    self.expr,
                    item,
                    self.column_types[2].elemType
                ),
                self.column_types[1].elemType.elemType,
                df=None,
                column_name=None,
                index_type=self.column_types[0].elemType,
                index_name=self.column_names[0]
            )

    @property
    def loc(self):
        return DataFrameWeldLoc(self)

    def merge(self, df2):
        if not isinstance(df2, DataFrameWeldExpr):
            raise Exception("df2 must be of type DataFrameWeldExpr")

        keys_d1 = set(self.colindex_map.keys())
        keys_d2 = set(df2.colindex_map.keys())
        join_keys = keys_d1 & keys_d2
        key_index_d1 = [self.colindex_map[key] for key in join_keys]
        key_index_d2 = [df2.colindex_map[key] for key in join_keys]
        rest_keys_d1 = keys_d1.difference(join_keys)
        rest_keys_d2 = keys_d2.difference(join_keys)
        rest_index_d1 = [self.colindex_map[key] for key in rest_keys_d1]
        rest_index_d2 = [df2.colindex_map[key] for key in rest_keys_d2]

        new_column_names = list(join_keys) + list(rest_keys_d1) + list(rest_keys_d2)

        # We add the key column first, followed by those in self, and finally df2

        key_index_types = []
        for i in key_index_d1:
            key_index_types.append(self.column_types[i])

        if len(key_index_types) > 1:
            join_keys_type = WeldStruct(key_index_types)
        else:
            join_keys_type = key_index_types[0]

        rest_types_d1 = []
        for i in rest_index_d1:
            rest_types_d1.append(self.column_types[i])

        rest_types_d2 = []
        for i in rest_index_d2:
            rest_types_d2.append(df2.column_types[i])

        new_types = key_index_types + rest_types_d1 + rest_types_d2
        return DataFrameWeldExpr(
            grizzly_impl.join(
                self.expr,
                df2.expr,
                key_index_d1,
                key_index_d2,
                join_keys_type,
                rest_index_d1,
                WeldStruct(rest_types_d1),
                rest_index_d2,
                WeldStruct(rest_types_d2)
            ),
            new_column_names,
            WeldStruct(new_types)
        )

    def pivot_table(self, values, index, columns, aggfunc='sum'):
        value_index = self.colindex_map[values]
        index_index = self.colindex_map[index]
        columns_index = self.colindex_map[columns]

        ind_ty = to_weld_type(self.column_types[index_index], 1)
        piv_ty = to_weld_type(WeldDouble(), 2)
        col_ty = to_weld_type(self.column_types[columns_index], 1)
        return DataFrameWeldExpr(
            grizzly_impl.pivot_table(
                self.expr,
                value_index,
                self.column_types[value_index],
                index_index,
                self.column_types[index_index],
                columns_index,
                self.column_types[columns_index],
                aggfunc
            ),
            [index, columns, values],
            WeldStruct([ind_ty, piv_ty, col_ty]),
            is_pivot=True
        )

    def sum(self, axis):
        if axis == 1:
            if self.is_pivot:
                if isinstance(self.column_types[1], WeldVec):
                    elem_type = self.column_types[1].elemType
                    value_type = elem_type.elemType
                    return SeriesWeld(
                        grizzly_impl.pivot_sum(
                            self.expr,
                            value_type
                        ),
                        value_type
                    )
            else:
                raise Exception("Sum for non-pivot table data frames not supported")
        elif axis == 0:
            raise Exception("Sum not implemented yet for axis = 0")

    def div(self, series, axis):
        if axis == 0:
            if self.is_pivot:
                if isinstance(self.column_types[1], WeldVec):
                    elem_type = self.column_types[1].elemType
                    value_type = elem_type.elemType
                    ind_ty = self.column_types[0]
                    piv_ty = to_weld_type(WeldDouble(), 2)
                    col_ty = self.column_types[2]
                    return DataFrameWeldExpr(
                        grizzly_impl.pivot_div(
                            self.expr,
                            series.expr,
                            elem_type,
                            value_type,
                        ),
                        self.column_names,
                        WeldStruct([ind_ty, piv_ty, col_ty]),
                        is_pivot=True
                    )
            else:
                raise Exception("Div for non-pivot table data frames not supported")
        elif axis == 1:
            raise Exception("Div not implemented yet for axis = 0")

    def sort_values(self, by):
        if self.is_pivot:
            return DataFrameWeldExpr(
                grizzly_impl.pivot_sort(
                    self.expr,
                    by,
                    self.column_types[0].elemType, # The index type
                    self.column_types[2].elemType, # The column name types (usually string)
                    self.column_types[1].elemType.elemType # The column value type
                ),
                self.column_names,
                self.weld_type,
                is_pivot=True
            )
        else:
            raise Expcetion("sort_values needs to be implemented for non pivot tables")

    def evaluate(self, verbose=True, passes=None):
        """Summary

        Returns:
            TYPE: Description
        """
        if self.is_pivot:
            index, pivot, columns = LazyOpResult(
                self.expr,
                self.weld_type,
                0
            ).evaluate(verbose=verbose, passes=passes)
            df_dict = {}
            for i, column_name in enumerate(columns):
                df_dict[column_name] = pivot[i]
            return DataFrameWeld(pd.DataFrame(df_dict, index=index))
        else:
            df = pd.DataFrame(columns=[])
            weldvec_type_list = []
            for type in self.column_types:
                weldvec_type_list.append(WeldVec(type))

            columns = LazyOpResult(
                grizzly_impl.unzip_columns(
                    self.expr,
                    self.column_types
                ),
                WeldStruct(weldvec_type_list),
                0
            ).evaluate(verbose=verbose, passes=passes)

            for i, column_name in enumerate(self.column_names):
                df[column_name] = columns[i]

            return DataFrameWeld(df)

    def get_column(self, column_name, column_type, index, verbose=True):
        """Summary

        Args:
            column_name (TYPE): Description
            column_type (TYPE): Description
            index (TYPE): Description

        Returns:
            TYPE: Description
        """
        return LazyOpResult(
            grizzly_impl.get_column(
                self.expr,
                self.weld_type,
                index
            ),
            column_type,
            1
        ).evaluate(verbose=verbose)

def merge(df1, df2):
    if isinstance(df1, DataFrameWeld):
        # TODO need to passin filtering by predicates as well here
        tys = []
        for col_name, raw_column in df1.raw_columns.items():
            dtype = str(raw_column.dtype)
            if dtype == 'object' or dtype == '|S64':
                weld_type = WeldVec(WeldChar())
            else:
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            tys.append(weld_type)

        if len(tys) == 1:
            weld_type = tys[0]
        else:
            weld_type = WeldStruct(tys)

        df1 = DataFrameWeldExpr(
            grizzly_impl.zip_columns(
                df1.raw_columns.values()
            ),
            df1.raw_columns.keys(),
            weld_type
        )
    if isinstance(df2, DataFrameWeld):
        tys = []
        for col_name, raw_column in df2.raw_columns.items():
            dtype = str(raw_column.dtype)
            if dtype == 'object' or dtype == '|S64':
                weld_type = WeldVec(WeldChar())
            else:
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            tys.append(weld_type)

        if len(tys) == 1:
            weld_type = tys[0]
        else:
            weld_type = WeldStruct(tys)
        df2 = DataFrameWeldExpr(
            grizzly_impl.zip_columns(
                df2.raw_columns.values()
            ),
            df2.raw_columns.keys(),
            weld_type
        )
    return df1.merge(df2)

def group_eval(objs, passes=None):
    LazyOpResults = []
    for ob in objs:
        if isinstance(ob, SeriesWeld):
            if ob.index_type is not None:
                weld_type = WeldStruct([WeldVec(ob.index_type), WeldVec(ob.weld_type)])
                LazyOpResults.append(LazyOpResult(ob.expr, weld_type, 0))
        else:
            LazyOpResults.append(LazyOpResult(ob.expr, ob.weld_type, 0))
    
    results = group(LazyOpResults).evaluate((True, -1), passes=passes)
    pd_results = []
    for i, result in enumerate(results):
        ob = objs[i]
        if isinstance(ob, SeriesWeld):
            if ob.index_type is not None:
                index, column = result
                series = pd.Series(column, index)
                series.index.rename(ob.index_name, True)
                pd_results.append(series)
            else:
                pd_results.append(series)
        if isinstance(ob, DataFrameWeldExpr):
            if ob.is_pivot:
                index, pivot, columns = result
                df_dict = {}
                for i, column_name in enumerate(columns):
                    df_dict[column_name] = pivot[i]
                pd_results.append(pd.DataFrame(df_dict, index=index))
            else:
                columns = result
                df_dict = {}
                for i, column_name in enumerate(ob.column_names):
                    df_dict[column_name] = columns[i]
                pd_results.append(pd.DataFrame(df_dict))
    return pd_results

def group(exprs):
    weld_type = [to_weld_type(expr.weld_type, expr.dim) for expr in exprs]
    exprs = [expr.expr for expr in exprs]
    weld_obj = WeldObject(grizzly_impl.encoder_, grizzly_impl.decoder_)
    weld_type = WeldStruct(weld_type)
    dim = 0

    expr_names = [expr.obj_id for expr in exprs]
    for expr in exprs:
        weld_obj.update(expr)
    weld_obj.weld_code = "{%s}" % ", ".join(expr_names)
    for expr in exprs:
        weld_obj.dependencies[expr.obj_id] = expr

    return LazyOpResult(weld_obj, weld_type, dim)


class DataFrameWeld:
    """Summary

    Attributes:
        df (TYPE): Description
        predicates (TYPE): Description
        unmaterialized_cols (TYPE): Description
        expr (TYPE): Description
    """

    def __init__(self, df, predicates=None, expr=None):
        self.df = df
        self.unmaterialized_cols = dict()
        self.predicates = predicates
        self.raw_columns = dict()
        for key in self.df:
            raw_column = self.df[key].values
            if raw_column.dtype == object:
                raw_column = np.array(self.df[key], dtype=str)
            self.raw_columns[key] = raw_column

    def __getitem__(self, key):
        """Summary

        Args:
            key (TYPE): Description

        Returns:
            TYPE: Description

        Raises:
            Exception: Description
        """
        if isinstance(key, str):  # Single-key get
            # First check if key corresponds to an un-materialized column
            if key in self.unmaterialized_cols:
                return self.unmaterialized_cols[key]
            raw_column = self.df[key].values
            dtype = str(raw_column.dtype)
            # If column type is "object", then cast as "vec[char]" in Weld
            if dtype == 'object':
                raw_column = self.raw_columns[key]
                weld_type = WeldVec(WeldChar())
            else:
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            if self.predicates is None:
                return SeriesWeld(raw_column, weld_type, self, key)
            return SeriesWeld(
                grizzly_impl.filter(
                    raw_column,
                    self.predicates.expr,
                    weld_type
                ),
                weld_type,
                self,
                key
            )
        elif isinstance(key, list):
            # For multi-key get, return type is a dataframe
            return DataFrameWeld(self.df[key], self.predicates)
        elif isinstance(key, SeriesWeld):
            # Can also apply predicate to a dataframe
            if self.predicates is not None:
                return DataFrameWeld(self.df, key.per_element_and(self.predicates))
            return DataFrameWeld(self.df, key)
        raise Exception("Invalid type in __getitem__")

    def __setitem__(self, key, value):
        """Summary

        Args:
            key (TYPE): Description
            value (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(value, np.ndarray):
            dtype = str(value.dtype)
            weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            self.unmaterialized_cols[key] = SeriesWeld(
                value,
                weld_type,
                self,
                key
            )
        elif isinstance(value, SeriesWeld):
            self.unmaterialized_cols[key] = value
        elif isinstance(value, LazyOpResult):
            self.unmaterialized_cols[key] = SeriesWeld(
                value.expr,
                value.weld_type,
                self,
                key
            )

    @property
    def values(self):
        if self.predicates is None:
            return self.df.values
        else:
            if isinstance(self.df.values, np.ndarray):
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[
                    str(self.df.values.dtype)]
                dim = self.df.values.ndim
                return LazyOpResult(
                    grizzly_impl.filter(
                        self.df.values,
                        self.predicates.expr,
                        weld_type
                    ),
                    weld_type,
                    dim
                )

    def _get_column_names(self):
        """Summary

        Returns:
            TYPE: Description
        """
        column_names = set()
        for column in self.df:
            column_names.add(column)
        for column in self.unmaterialized_cols:
            column_names.add(column)
        return list(column_names)

    def filter(self, predicates):
        """Summary

        Args:
            grouping_column_name (TYPE): Description

        Returns:
            TYPE: Description
        """
        tys = []
        for col_name, raw_column in self.raw_columns.items():
            dtype = str(raw_column.dtype)
            if dtype == 'object' or dtype == '|S64':
                weld_type = WeldVec(WeldChar())
            else:
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            tys.append(weld_type)

        if len(tys) == 1:
            weld_type = tys[0]
        else:
            weld_type = WeldStruct(tys)

        if isinstance(predicates, SeriesWeld):
            predicates = predicates.expr

        return DataFrameWeldExpr(
            grizzly_impl.filter(
                grizzly_impl.zip_columns(
                    self.raw_columns.values(),
                ),
                predicates
            ),
            self.raw_columns.keys(),
            weld_type
        )

    def pivot_table(self, values, index, columns, aggfunc='sum'):
        tys = []
        for col_name, raw_column in self.raw_columns.items():
            dtype = str(raw_column.dtype)
            if dtype == 'object' or dtype == '|S64':
                weld_type = WeldVec(WeldChar())
            else:
                weld_type = grizzly_impl.numpy_to_weld_type_mapping[dtype]
            tys.append(weld_type)

        if len(tys) == 1:
            weld_type = tys[0]
        else:
            weld_type = WeldStruct(tys)

        return DataFrameWeldExpr(
            grizzly_impl.zip_columns(
                self.raw_columns.values(),
            ),
            self.raw_columns.keys(),
            weld_type
        ).pivot_table(values, index, columns, aggfunc)

    def groupby(self, grouping_column_name):
        """Summary

        Args:
            grouping_column_name (TYPE): Description

        Returns:
            TYPE: Description
        """
        return GroupByWeld(
            self,
            grouping_column_name
        )

    def to_pandas(self):
        """Summary

        Returns:
            TYPE: Description
        """
        # TODO: Do more work here (need to materialize all columns as needed)
        return self.df


class GroupedDataFrameWeld(LazyOpResult):
    """Summary

    Attributes:
        column_names (TYPE): Description
        column_types (TYPE): Description
        dim (int): Description
        expr (TYPE): Description
        grouping_column_name (TYPE): Description
        grouping_column_type (TYPE): Description
        weld_type (TYPE): Description
        ptr (TYPE): Description
    """

    def __init__(
            self,
            expr,
            grouping_column_names,
            column_names,
            grouping_column_types,
            column_types):
        """Summary

        Args:
            expr (TYPE): Description
            grouping_column_name (TYPE): Description
            column_names (TYPE): Description
            grouping_column_type (TYPE): Description
            column_types (TYPE): Description
        """
        self.expr = expr
        self.grouping_column_name = grouping_column_names
        self.column_names = column_names
        self.grouping_column_types = grouping_column_types
        self.column_types = column_types
        self.ptr = None

        self.dim = 1
        if isinstance(self.column_types, list):
            if len(self.column_types) == 1:
                column_types = self.column_types[0]
            else:
                column_types = WeldStruct(self.column_types)

        if len(self.grouping_column_types) == 1:
            grouping_column_types = self.grouping_column_types[0]
        else:
            grouping_column_types = WeldStruct(self.grouping_column_types)
        self.weld_type = WeldStruct([grouping_column_types, column_types])

    def slice(self, start, size):
        return GroupedDataFrameWeld(
            grizzly_impl.grouped_slice(
                self.expr,
                self.weld_type,
                start,
                size
            ),
            self.grouping_column_name,
            self.column_names,
            self.grouping_column_types,
            self.column_types
        )

    def get_column(self, column_name, column_type, index, verbose=True):
        """Summary

        Args:
            column_name (TYPE): Description
            column_type (TYPE): Description
            index (TYPE): Description

        Returns:
            TYPE: Description
        """
        return LazyOpResult(
            grizzly_impl.get_column(
                self.expr,
                self.weld_type,
                index
            ),
            column_type,
            1
        )

    def reset_index(self, inplace=True, drop=True):
        """ Flattens the grouped data structure.

        #TODO: The parameters here are meaningless

        Flattens the grouped data structure.
        What is returned is a DataFrameWeld object.
        """
        if len(self.column_types) == 1:
            vectype = self.column_types[0]
            if isinstance(vectype, WeldVec):
                elem_type = vectype.elemType
                if isinstance(elem_type, WeldStruct):
                    self.column_types = elem_type.field_types
                    value_type = WeldStruct(self.column_types)
                else:
                    self.column_types = elem_type
                    value_type = elem_type
                if len(self.grouping_column_types) == 1:
                    group_type = self.grouping_column_types[0]
                else:
                    group_type = WeldStruct(self.grouping_column_types)

                self.weld_type = WeldStruct([group_type, value_type])
                self.expr = grizzly_impl.flatten_group(
                    self.expr,
                    self.column_types,
                    self.grouping_column_types
                )

    def evaluate(self, verbose=True, passes=None):
        """Summary

        Returns:
            TYPE: Description
        """
        i = 0
        exprs = []
        for column_name in self.column_names:
            if len(self.column_names) > 1:
                index = "1.$%d" % i
            else:
                index = "1"
            expr = self.get_column(
                column_name,
                self.column_types[i],
                index,
                verbose=verbose
            )
            i += 1
            exprs.append(expr)

        i = 0
        for column_name in self.grouping_column_name:
            if len(self.grouping_column_name) > 1:
                index = "0.$%d" % i
            else:
                index = "0"
            expr = self.get_column(
                column_name,
                self.grouping_column_types[i],
                index,
                verbose=verbose
            )
            exprs.append(expr)
            i += 1

        result = group(exprs).evaluate(verbose=True, passes=passes)
        df = pd.DataFrame(columns=[])
        all_columns = self.column_names + self.grouping_column_name
        for i, column_name in enumerate(all_columns):
            df[column_name] = result[i]
        return DataFrameWeld(df)

class SeriesWeld(LazyOpResult):
    """Summary

    Attributes:
        column_name (TYPE): Description
        df (TYPE): Description
        dim (int): Description
        expr (TYPE): Description
        weld_type (TYPE): Description
    """

    def __init__(self, expr, weld_type, df=None, column_name=None, index_type=None, index_name=None):
        """Summary

        TODO: Implement an actual Index Object like how Pandas does
        Args:
            expr (TYPE): Description
            weld_type (TYPE): Description
            df (None, optional): Description
            column_name (None, optional): Description
        """
        self.expr = expr
        self.weld_type = weld_type
        self.dim = 1
        self.df = df
        self.column_name = column_name
        self.index_type = index_type
        self.index_name = index_name

    def __getitem__(self, key):
        """Summary

        Args:
            predicates (TYPE): Description
            new_value (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(key, slice):
            start = key.start
            # TODO : We currently do nothing with step
            step = key.step
            stop = key.stop
            if self.index_type is not None:
                index_expr = grizzly_impl.get_field(self.expr, 0)
                column_expr = grizzly_impl.get_field(self.expr, 1)
                zip_expr = grizzly_impl.zip_columns([index_expr, column_expr])
                sliced_expr = grizzly_impl.slice_vec(zip_expr, start, stop)
                unzip_expr = grizzly_impl.unzip_columns(
                    sliced_expr,
                    [self.index_type, self.weld_type]
                )
                return SeriesWeld(
                    unzip_expr,
                    self.weld_type,
                    self.df,
                    self.column_name,
                    self.index_type,
                    self.index_name
                )
            else:
                return SeriesWeld(
                    grizzly_impl.slice_vec(
                        self.expr,
                        start,
                        stop
                    )
                )
        else:
            # By default we return as if the key were predicates to filter by
            return self.filter(key)

    def __setitem__(self, predicates, new_value):
        """Summary

        Args:
            predicates (TYPE): Description
            new_value (TYPE): Description

        Returns:
            TYPE: Description
        """
        if self.df is not None and self.column_name is not None:
            self.df[self.column_name] = self.mask(predicates, new_value)

    @property
    def loc(self):
        return WeldLocIndexer(
            self
        )

    def __getattr__(self, key):
        """Summary

        Args:
            key (TYPE): Description

        Returns:
            TYPE: Description

        Raises:
            Exception: Description
        """
        if key == 'str' and self.weld_type == WeldVec(WeldChar()):
            return StringSeriesWeld(
                self.expr,
                self.weld_type,
                self.df,
                self.column_name
            )
        raise AttributeError("Attr %s does not exist" % key)

    @property
    def index(self):
        if self.index_type is not None:
            return SeriesWeld(
                grizzly_impl.get_field(
                    self.expr,
                    0
                ),
                self.index_type,
                self.df,
                self.index_name
            )
        # TODO : Make all series have a series attribute
        raise Exception("No index present")

    def evaluate(self, verbose=False, passes=None):
        if self.index_type is not None:
            index, column = LazyOpResult(
                self.expr,
                WeldStruct([WeldVec(self.index_type), WeldVec(self.weld_type)]),
                0
            ).evaluate(verbose=verbose, passes=passes)
            series = pd.Series(column, index)
            series.index.rename(self.index_name, True)
            return series
        else:
            column = LazyOpResult.evaluate(self, verbose=verbose, passes=passes)
            return pd.Series(column)

    def sort_values(self, ascending=False):
        """ Sorts the values of this series

        """
        if self.index_type is not None:
            index_expr = grizzly_impl.get_field(self.expr, 0)
            column_expr = grizzly_impl.get_field(self.expr, 1)
            zip_expr = grizzly_impl.zip_columns([index_expr, column_expr])
            result_expr = grizzly_impl.sort(zip_expr, 1, self.weld_type, ascending)
            unzip_expr = grizzly_impl.unzip_columns(
                result_expr,
                [self.index_type, self.weld_type]
            )
            return SeriesWeld(
                unzip_expr,
                self.weld_type,
                self.df,
                self.column_name,
                self.index_type,
                self.index_name
            )
        else:
            result_expr = grizzly_impl.sort(self.expr)
            # TODO need to finish this

    def unique(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.unique(
                self.expr,
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def lower(self):
        """Summary

        Returns:
            TYPE: Description
        """
        # TODO : Bug in nested map operating on strings
        # TODO : Check that self.weld_type is a string type
        vectype = self.weld_type
        if isinstance(vectype, WeldVec):
            elem_type = vectype.elemType
            if isinstance(elem_type, WeldChar):
                        return SeriesWeld(
                            grizzly_impl.to_lower(
                                self.expr,
                                elem_type
                            ),
                            self.weld_type,
                            self.df,
                            self.column_name
                        )
        raise Exception("Cannot call to_lower on non string type")

    def contains(self, string):
        """Summary

        Returns:
        TYPE: Description
        """
        # Check that self.weld_type is a string type
        vectype = self.weld_type
        if isinstance(vectype, WeldVec):
            elem_type = vectype.elemType
            if isinstance(elem_type, WeldChar):
                return SeriesWeld(
                    grizzly_impl.contains(
                        self.expr,
                        elem_type,
                        string
                    ),
                    WeldBit(),
                    self.df,
                    self.column_name
                )
        raise Exception("Cannot call to_lower on non string type")

    def isin(self, ls):
        if isinstance(ls, SeriesWeld):
            if self.weld_type == ls.weld_type:
                return SeriesWeld(
                    grizzly_impl.isin(self.expr,
                                      ls.expr,
                                      self.weld_type),
                    WeldBit(),
                    self.df,
                    self.column_name
                )
        raise Exception("Cannot call isin on different typed list")

    def prod(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return LazyOpResult(
            grizzly_impl.aggr(
                self.expr,
                "*",
                1,
                self.weld_type
            ),
            self.weld_type,
            0
        )

    def sum(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return LazyOpResult(
            grizzly_impl.aggr(
                self.expr,
                "+",
                0,
                self.weld_type
            ),
            self.weld_type,
            0
        )

    def max(self):
        """Summary

        Returns:
            TYPE: Description
        """
        pass

    def min(self):
        """Summary

        Returns:
            TYPE: Description
        """
        pass

    def count(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return LazyOpResult(
            grizzly_impl.count(
                self.expr,
                self.weld_type
            ),
            WeldInt(),
            0
        )

    def mask(self, predicates, new_value):
        """Summary

        Args:
            predicates (TYPE): Description
            new_value (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(predicates, SeriesWeld):
            predicates = predicates.expr
        return SeriesWeld(
            grizzly_impl.mask(
                self.expr,
                predicates,
                new_value,
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def filter(self, predicates):
        if isinstance(predicates, SeriesWeld):
            predicates = predicates.expr
        return SeriesWeld(
            grizzly_impl.filter(
                self.expr,
                predicates,
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def add(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "+",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def __sub__(self, other):
        # TODO subtractionw without index variables
        if self.index_type is not None:
            index = grizzly_impl.get_field(self.expr, 0)
            expr1 = grizzly_impl.get_field(self.expr, 1)
        else:
            expr1 = self.expr
        if other.index_type is not None:
            index2 = grizzly_impl.get_field(other.expr, 0)
            expr2 = grizzly_impl.get_field(other.expr, 1)
        else:
            expr2 = other.expr
        index_expr = LazyOpResult(index, self.index_type, 0)
        sub_expr = SeriesWeld(
            grizzly_impl.element_wise_op(
                expr1,
                expr2,
                "-",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

        index_sub_expr = group([index_expr, sub_expr])
        return SeriesWeld(
            index_sub_expr.expr,
            self.weld_type,
            self.df,
            self.column_name,
            self.index_type,
            self.index_name
        )
        # We also need to ensure that both indexes of the subtracted
        # columns are compatible

    def sub(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "-",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def mul(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "*",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def div(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "/",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def per_element_and(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "&&",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def mod(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if isinstance(other, SeriesWeld):
            other = other.expr
        return SeriesWeld(
            grizzly_impl.element_wise_op(
                self.expr,
                other,
                "%",
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

    def __eq__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.compare(
                self.expr,
                other,
                "==",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )

    def __ne__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.compare(
                self.expr,
                other,
                "!=",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )

    def __gt__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.compare(
                self.expr,
                other,
                ">",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )

    def __ge__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        if self.index_type is not None:
            expr = grizzly_impl.get_field(self.expr, 1)
        else:
            expr = self.expr
        return SeriesWeld(
            grizzly_impl.compare(
                expr,
                other,
                ">=",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )

    def __lt__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.compare(
                self.expr,
                other,
                "<",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )

    def __le__(self, other):
        """Summary

        Args:
            other (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.compare(
                self.expr,
                other,
                "<=",
                self.weld_type
            ),
            WeldBit(),
            self.df,
            self.column_name
        )


class StringSeriesWeld:
    """Summary

    Attributes:
        column_name (TYPE): Description
        df (TYPE): Description
        dim (int): Description
        expr (TYPE): Description
        weld_type (TYPE): Description
    """

    def __init__(self, expr, weld_type, df=None, column_name=None):
        """Summary

        Args:
            expr (TYPE): Description
            weld_type (TYPE): Description
            df (None, optional): Description
            column_name (None, optional): Description
        """
        self.expr = expr
        self.weld_type = weld_type
        self.dim = 1
        self.df = df
        self.column_name = column_name

    def slice(self, start, size):
        """Summary

        Args:
            start (TYPE): Description
            size (TYPE): Description

        Returns:
            TYPE: Description
        """
        return SeriesWeld(
            grizzly_impl.slice(
                self.expr,
                start,
                size,
                self.weld_type
            ),
            self.weld_type,
            self.df,
            self.column_name
        )

class GroupByWeldSeries:
    """Summary

    Attributes:
        column_name
        column_type
        column,
        grouping_column
    """

    def __init__(self, name, column, column_type, grouping_column_names, grouping_columns, grouping_column_types):
        self.name = name
        self.column = column
        self.column_type = column_type
        self.grouping_column_names = grouping_column_names
        self.grouping_columns = grouping_columns
        self.grouping_column_types = grouping_column_types

    def std(self):
        """Standard deviation

        Note that is by default normalizd by n - 1
        # TODO, what does pandas do for multiple grouping columns?
        # Currently we are just going to use one grouping column
        """
        std_expr = grizzly_impl.groupby_std(
            [self.column],
            [self.column_type],
            self.grouping_columns,
            self.grouping_column_types
        )
        unzipped_columns = grizzly_impl.unzip_columns(
            std_expr,
            self.grouping_column_types + [WeldDouble()],
        )
        index_expr = LazyOpResult(
            grizzly_impl.get_field(unzipped_columns, 0),
            self.grouping_column_types[0],
            1
        )
        column_expr = LazyOpResult(
            grizzly_impl.get_field(unzipped_columns, 1),
            self.grouping_column_types[0],
            1
        )
        group_expr = group([index_expr, column_expr])
        return SeriesWeld(
            group_expr.expr,
            WeldDouble(),
            index_type=self.grouping_column_types[0],
            index_name=self.grouping_column_names[0]
        )

class GroupByWeld:
    """Summary

    Attributes:
        column_names (TYPE): Description
        column_types (list): Description
        columns (list): Description
        grouping_column (TYPE): Description
        grouping_column_name (TYPE): Description
        grouping_column_type (TYPE): Description
    """

    def __init__(self, df, grouping_column_names):
        """Summary

        Args:
            df (TYPE): Description
            grouping_column_name (TYPE): Description
        """
        self.df = df
        self.grouping_columns = []
        self.grouping_column_types = []

        if isinstance(grouping_column_names, str):
            grouping_column_names = [grouping_column_names]
        for column_name in grouping_column_names:
            column = df[column_name]
            if isinstance(column, LazyOpResult):
                self.grouping_column_types.append(column.weld_type)
                self.grouping_columns.append(column.expr)
            elif isinstance(column, np.ndarray):
                column_type = numpyImpl.numpy_to_weld_type_mapping[
                    str(column.dtype)]
                self.grouping_column_types.append(column_type)
                self.grouping_columns.append(column)

        self.grouping_column_names = grouping_column_names
        self.column_names = []
        for x in df._get_column_names():
            if x not in self.grouping_column_names:
                self.column_names.append(x)

        self.columns = []
        self.column_types = []
        for column_name in self.column_names:
            column = df[column_name]
            column_type = None
            if isinstance(column, LazyOpResult):
                column_type = column.weld_type
                column = column.expr
            elif isinstance(column, np.ndarray):
                column_type = numpyImpl.numpy_to_weld_type_mapping[
                    str(column.dtype)]

            self.columns.append(column)
            self.column_types.append(column_type)

    def __getitem__(self, item):
        item_index = self.column_names.index(item)
        return GroupByWeldSeries(
            item,
            self.columns[item_index],
            self.column_types[item_index],
            self.grouping_column_names,
            self.grouping_columns,
            self.grouping_column_types
        )

    def sum(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return GroupedDataFrameWeld(
            grizzly_impl.groupby_sum(
                self.columns,
                self.column_types,
                self.grouping_columns,
                self.grouping_column_types
            ),
            self.grouping_column_names,
            self.column_names,
            self.grouping_column_types,
            self.column_types
        )

    def sort_values(self, by, ascending=True):
        """Summary

        Returns:
            TYPE: Description
        """
        if len(self.column_types) == 1:
            vec_type = [WeldVec(self.column_types[0])]
        else:
            vec_type = [WeldVec(WeldStruct(self.column_types))]

        if len(self.column_names) > 1:
            key_index = self.column_names.index(by)
        else :
            key_index = None

        return GroupedDataFrameWeld(
            grizzly_impl.groupby_sort(
                self.columns,
                self.column_types,
                self.grouping_columns,
                self.grouping_column_types,
                key_index,
                ascending
            ),
            self.grouping_column_names,
            self.column_names,
            self.grouping_column_types,
            vec_type
        )

    def mean(self):
        """Summary

        Returns:
            TYPE: Description
        """
        pass

    def size(self):
        """Returns the sizes of the groups as series.

        Returns:
            TYPE: Description
        """
        if len(self.grouping_column_types) > 1:
            index_type = WeldStruct([self.grouping_column_types])
            # Figure out what to use for multi-key index name
            # index_name = ??
        else:
            index_type = self.grouping_column_types[0]
            index_name = self.grouping_column_names[0]
        return SeriesWeld(
            grizzly_impl.groupby_size(
                self.columns,
                self.column_types,
                self.grouping_columns,
                self.grouping_column_types
            ),
            WeldLong(),
            index_type=index_type,
            index_name=index_name
        )

    def count(self):
        """Summary

        Returns:
            TYPE: Description
        """
        pass

    def apply(self, func):
        return func(self)



import os
import platform
import shutil
import subprocess
import sys

from setuptools import setup, Distribution
import setuptools.command.build_ext as _build_ext
from setuptools.command.install import install
from setuptools.command.develop import develop

system = platform.system()
if system == 'Linux':
    numpy_convertor = "numpy_weld_convertor.so"
elif system == 'Windows':
    numpy_convertor = "numpy_weld_convertor.dll"
elif system == 'Darwin':
    numpy_convertor = "numpy_weld_convertor.dylib"
else:
    raise OSError("Unsupported platform {}", system)

class Install(install):
    def run(self):
        install.run(self)
        python_executable = sys.executable
        protoc_command_clean = ["make -C " + self.install_lib + "grizzly clean"]
        if subprocess.call(protoc_command_clean, shell=True) != 0:
            sys.exit(-1)

        protoc_command_make = ["make -C " + self.install_lib + "grizzly/ EXEC=" + python_executable]
        if subprocess.call(protoc_command_make, shell=True) != 0:
            sys.exit(-1)

class Develop(develop):
    def run(self):
        develop.run(self)
        python_executable = sys.executable
        protoc_command_clean = ["make -C grizzly/ clean"]
        if subprocess.call(protoc_command_clean, shell=True) != 0:
            sys.exit(-1)

        protoc_command_make = ["make -C grizzly/ EXEC=" + python_executable]
        if subprocess.call(protoc_command_make, shell=True) != 0:
            sys.exit(-1)

class BinaryDistribution(Distribution):
    def has_ext_modules(self):
        return True

setup(name='pygrizzly',
      version='0.0.2',
      packages=['grizzly'],
      package_data = {'grizzly': ['numpy_weld_convertor.cpp', 'common.h', 'Makefile']},
      cmdclass={"install": Install, "develop": Develop},
      distclass=BinaryDistribution,
      url='https://github.com/weld-project/weld',
      author='Weld Developers',
      author_email='weld-group@lists.stanford.edu',
      install_requires=['pyweld'])
